#define PI_NO_NSEND
#define PI_NO_NRECV
#define MPID_HAS_HETERO

/*
 *  $Id: chsync.c,v 1.6 1994/08/12 20:39:53 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vc[] = "$Id: chsync.c,v 1.6 1994/08/12 20:39:53 gropp Exp $";
#endif

#include "mpid.h"


static int DebugFlag = 0;

void MPID_SetSyncDebugFlag( f )
int f;
{
DebugFlag = f;
}

/*
   This file contains routines to keep track of synchronous send messages;
   they must be explicitly acknowledged.
 */

typedef struct _MPID_SyncId {
    int id;
    MPIR_SHANDLE *dmpi_send_handle;
    MPID_SHANDLE *mpid_send_handle;
    struct _MPID_SyncId *next;
    } MPID_SyncId;

static MPID_SyncId *head = 0;
static int         CurId = 1;
/* Get a message id for this message and add to the list */
/* 
   KNOWN BUG: With a huge number of messages, we could eventually get rollover
   in the CurId; to manage this, we would need to send informational messages
   to our friends occasionally
 */

void MPID_P4_Get_Sync_Id( pkt, dmpi_handle, mpid_handle )
MPID_PACKET  *pkt;
MPIR_SHANDLE *dmpi_handle;
MPID_SHANDLE *mpid_handle;
{
MPID_SyncId *new;

MPID_SYNC_SET_ID(pkt,CurId);
new		      = (MPID_SyncId *)malloc(sizeof(MPID_SyncId));    /* This should use SBalloc code */
if (!new) MPID_P4_Abort( MPI_ERR_EXHAUSTED );
new->id		      = CurId++;
new->dmpi_send_handle = dmpi_handle;
new->mpid_send_handle = mpid_handle;
new->next	      = head;
head		      = new;
}

int MPID_P4_Lookup_SyncAck( pkt, dmpi_send_handle, mpid_send_handle )
MPID_PACKET  *pkt;
MPIR_SHANDLE **dmpi_send_handle;
MPID_SHANDLE **mpid_send_handle;
{
MPID_SyncId *cur, *last;
int         modeid = MPID_SYNC_GET_ID(pkt);

last		  = 0;
cur		  = head;
*dmpi_send_handle = 0;
*mpid_send_handle = 0;
while (cur) {
    if (cur->id == modeid) {
	/* Note that the fields may be NULL if this message was cancelled */
	*dmpi_send_handle = cur->dmpi_send_handle;
	*mpid_send_handle = cur->mpid_send_handle;
	if (last) last->next = cur->next;
	else      head       = cur->next;
	/* See if we can easily recover the id */
	if (CurId == modeid + 1) CurId--;
	free(cur );
	return;
	}
    last = cur;
    cur  = cur->next;
    }
/* Error, did not find id! */
if (!dmpi_send_handle) {
    fprintf( stderr, "Error in processing sync id %x!\n", modeid );
    }
}

/* Process a synchronization acknowledgement */
MPID_SyncAck( pkt, from )
MPID_PACKET *pkt;
int         from;
{
MPIR_SHANDLE *dmpi_send_handle;
MPID_SHANDLE *mpid_send_handle;

/* This is an acknowledgement of a synchronous send; look it up and
   mark as completed */
if (DebugFlag) {
    printf( 
    "[%d]SYNC received sync ack message for mode=%x from %d (%s:%d)\n",  
	   __MYPROCID, MPID_SYNC_GET_ID(pkt), from, __FILE__, __LINE__ );
    fflush( stdout );
    }
MPID_P4_Lookup_SyncAck( pkt, &dmpi_send_handle, &mpid_send_handle );
if (!dmpi_send_handle) {
    fprintf( stderr, "Error in processing sync ack!\n" );
    return MPI_ERR_INTERN;
    }
dmpi_send_handle->completed = 1;
mpid_send_handle->done	    = 1;
}

/* return an acknowledgment */
MPID_SyncReturnAck( sync_id, from )
int sync_id, from;
{
MPID_PACKET pkt;

/* We clear the entries of the packet so that debugging will be easier... */
pkt.tag	       = 0;
pkt.len	       = 0;
pkt.context_id = 0;
pkt.lrank      = 0;
pkt.mode       = MPID_MODE_SYNC_ACK;
MPID_SYNC_SET_ID(&pkt,sync_id);

/* Rest of packet is irrelavent */
if (DebugFlag) {
    printf( 
   "[%d]SYNC Starting a send of tag = %d, dest = %d, mode=",
	    __MYPROCID, MPID_PT2PT_TAG, from );
    MPID_Print_mode( stdout, &pkt );
    fprintf( stdout, "(%s:%d)\n", __FILE__, __LINE__ );
    MPID_Print_packet( stdout, &pkt );
    fflush( stdout );
    }
p4_sendx(MPID_PT2PT_TAG,from,(char *)(&pkt),MPID_HEADER_LEN,P4NOX );
}

/* Look through entire list for this API handle */
MPID_Sync_discard( dmpi )
MPIR_SHANDLE *dmpi;
{
MPID_SyncId *cur, *last;

last		  = 0;
cur		  = head;
while (cur) {
    if (cur->dmpi_send_handle == dmpi) {
	cur->dmpi_send_handle = 0;
 	cur->mpid_send_handle = 0;
	/* We leave this here just in case a reply is received, so that 
	   we can distinquish between invalid syncAck messages and
	   syncAcks for cancelled messages.  */
	return;
	}
    last = cur;
    cur  = cur->next;
    }
/* Error, did not find id! Ignored for now. */
}
